<?php

namespace Ardent\Undercurrent\Container;

use Exception;
use ReflectionClass;

class GenericContainer implements ContainerInterface
{
    private array $definitions = [];

    private array $instances = [];

    private array $aliases = [];

    public function add(string $className, ?callable $definition = null, bool $singleton = true): self
    {
        if (!$definition) {
            $definition = fn() => $this->autowire($className);
        }

        $this->definitions[$className] = [
            'definition' => $definition,
            'singleton' => $singleton,
        ];

        return $this;
    }

    public function alias(string $alias, string $className): self
    {
        $this->aliases[$alias] = $className;

        return $this;
    }

    /**
     * @inheritDoc
     */
    public function get(string $className): object
    {
        if (isset($this->aliases[$className])) {
            $className = $this->aliases[$className];
        }
        if (!isset($this->definitions[$className])) {
            throw new ClassNotFoundException($className);
        }

        $definition = $this->definitions[$className]['definition'];

        if ($this->definitions[$className]['singleton']) {
            if (isset($this->instances[$className])) {
                return $this->instances[$className];
            }
            $instance = $definition($this);
            $this->instances[$className] = $instance;
        } else {
            $instance = $definition($this);
        }

        return $instance;
    }

    private function autowire(string $className): ?object
    {
        $reflection = new ReflectionClass($className);
        $constructor = $reflection->getConstructor();
        if (!$constructor) {
            return new $className();
        }

        $params = [];
        foreach ($constructor->getParameters() as $parameter) {
            $type = $parameter->getType();
            if (!$type) {
                throw new Exception("Parameter {$parameter->getName()} in $className has no type");
            }
            $params[] = $this->get($type->getName());
        }

        return new $className(...$params);
    }
}